/*
 *	$Id: fas.c
 */

/*
 * AMIGA Emulex FAS216 scsi adaptor driver
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/buf.h>
#include <sys/proc.h>
#include <scsi/scsi_all.h>
#include <scsi/scsiconf.h>
#include <vm/vm.h>
#include <vm/vm_kern.h>
#include <vm/vm_page.h>
#include <machine/pmap.h>
#include <machine/cpu.h>
#include <amiga/amiga/device.h>
#include <amiga/amiga/cc.h>
#include <amiga/amiga/custom.h>
#include <amiga/dev/fasreg.h>
#include <amiga/dev/fasvar.h>

void	fas_minphys __P((struct buf *bp));
int	fas_scsicmd __P((struct scsi_xfer *xs));

void	fasinitialize __P((struct fas_softc *));

int	fas_donextcmd(struct fas_softc *dev, struct fas_pending *pendp);
void	fas_scsidone(struct fas_softc *dev, struct scsi_xfer *xs, int stat);
void	fasintr(struct fas_softc *dev);

void	fasreset(struct fas_softc *dev, int how);
int	fasselect(struct fas_softc *dev, struct fas_pending *pendp,
		  unsigned char *cbuf, int clen, 
		  unsigned char *buf, int len, int mode);

#ifdef DEBUG
#define QPRINTF(a) if (fas_debug > 1) printf a
int	fas_debug = 0;
#else
#define QPRINTF
#endif

/*
 * default minphys routine for sci based controllers
 */
void fas_minphys(struct buf *bp)
{	/* no max transfer at this level */
}

void fas_init_nexus(struct fas_softc *dev, struct nexus *nexus)
{
  bzero(nexus, sizeof(struct nexus));

  nexus->state			= FAS_NS_IDLE;
  nexus->period			= 200;
  nexus->offset			= 0;
  nexus->syncper		= 5;
  nexus->syncoff		= 0;
  nexus->config3		= dev->sc_config3 & ~FAS_CFG3_FASTSCSI;
}

void fasinitialize(struct fas_softc *dev)
{
  fas_regmap_p	rp;
  int		i;

  dev->sc_led_status = 0;

  TAILQ_INIT(&dev->sc_xs_pending);
  TAILQ_INIT(&dev->sc_xs_free);

  for(i=0; i<MAXPENDING; i++)
    {
#ifdef FAS_NEED_VM_PATCH
      dev->sc_xs_store[i].vm_link_data.pages = 0;
#endif
      TAILQ_INSERT_TAIL(&dev->sc_xs_free, &dev->sc_xs_store[i], link);
    }

  if (dev->sc_clock_freq <= 10)
    dev->sc_clock_conv_fact = 2;
  else if (dev->sc_clock_freq <= 40)
    dev->sc_clock_conv_fact = 2+((dev->sc_clock_freq-10)/5);
  else
    {
      dev->sc_clock_conv_fact = 8;
      panic("fasinitialize: Clock frequence too high");
    }

  dev->sc_config1 = (dev->sc_host_id & FAS_CFG1_BUS_ID_MASK);
  dev->sc_config2 = FAS_CFG2_FEATURES_ENABLE;
  dev->sc_config3 = (dev->sc_clock_freq > 25 ? FAS_CFG3_FASTCLK : 0);

  dev->sc_timeout_val =
    1+dev->sc_timeout/(0.001/dev->sc_clock_freq*7682*dev->sc_clock_conv_fact);
  dev->sc_clock_period = 1000/dev->sc_clock_freq;

  fasreset(dev, 1 | 2);	/* Reset Chip and Bus */

  dev->sc_units_disconnected = 0;
  dev->sc_msg_in_len = 0;
  dev->sc_msg_out_len = 0;

  dev->sc_flags = 0;

  for(i=0; i<8; i++)
    fas_init_nexus(dev, &dev->sc_nexus[i]);

#ifdef FAS_NEED_VM_PATCH
  vm_map_lock(kernel_map);
  if (vm_map_findspace(kernel_map, 0, MAXPHYS+NBPG,
		       (vm_offset_t *)&dev->sc_vm_link))
    {
      vm_map_unlock(kernel_map);
      panic("FAS_SCSICMD: No VM space available.");
    }
  else
    {
      int	offset;
      
      offset = (vm_offset_t)dev->sc_vm_link - VM_MIN_KERNEL_ADDRESS;
      printf(" vm_link %x", dev->sc_vm_link);
      vm_object_reference(kernel_object);
      vm_map_insert(kernel_map, kernel_object, offset,
		    (vm_offset_t)dev->sc_vm_link,
		    (vm_offset_t)dev->sc_vm_link+(MAXPHYS+NBPG));
      vm_map_unlock(kernel_map);
    }

  dev->sc_vm_link_data.pages = 0;
  dev->sc_vm_link_data.offset = 0;
#endif
}

#ifdef FAS_NEED_VM_PATCH
void fas_unlink_vm_link(struct fas_softc *dev)
{
  if (dev->sc_flags & FAS_HAS_VM_LINK)
    {
      physunaccess(dev->sc_vm_link, dev->sc_vm_link_data.pages*NBPG);
      dev->sc_vm_link_data.pages = 0;
      dev->sc_vm_link_data.offset = 0;
      dev->sc_flags &= ~FAS_HAS_VM_LINK;
    }
}

void fas_link_vm_link(struct fas_softc *dev, struct vm_link_data *vm_link_data)
{
  int	i;

  if (dev->sc_flags & FAS_HAS_VM_LINK)
    fas_unlink_vm_link(dev);

  dev->sc_vm_link_data = *vm_link_data;

  if (dev->sc_vm_link_data.pages)
    {
      for(i=0; i<dev->sc_vm_link_data.pages; i++)
	physaccess(dev->sc_vm_link+i*NBPG, dev->sc_vm_link_data.pa[i],
		   NBPG, PG_CI);
      
      dev->sc_flags |= FAS_HAS_VM_LINK;
    }
}
#endif

/* used by specific scem controller */
int fas_scsicmd(struct scsi_xfer *xs)
{
  struct fas_softc	*dev;
  struct scsi_link	*slp;
  struct fas_pending	*pendp;
  int			 flags, s, target;
#ifdef FAS_NEED_VM_PATCH
  struct vm_link_data	 vm_link_data;
#endif

  slp = xs->sc_link;
  dev = slp->adapter_softc;
  flags = xs->flags;
  target = slp->target;

  if (flags & SCSI_DATA_UIO)
    panic("fas: scsi data uio requested");

  s = splbio();
  pendp = dev->sc_xs_free.tqh_first;
  if (!pendp)
    {
      splx(s);
      return(TRY_AGAIN_LATER);
    }
  TAILQ_REMOVE(&dev->sc_xs_free, pendp, link);
  pendp->xs = xs;

  splx(s);

  if ((flags & SCSI_POLL) && (dev->sc_flags & FAS_ACTIVE))
    panic("fas_scsicmd: busy");

#ifdef FAS_NEED_VM_PATCH
  pendp->vm_link_data.offset = 0;
  pendp->vm_link_data.pages = 0;

  if (!(flags & SCSI_POLL) && (
#ifdef M68040
			(cpu040 && ((vm_offset_t)xs->data >= 0xFFFC0000)) &&
#endif
			((vm_offset_t)xs->data >= 0xFF000000)))
    {
      vm_offset_t sva;
      short	  n;
	    
      sva = (vm_offset_t)xs->data & PG_FRAME;

      pendp->vm_link_data.offset = (vm_offset_t)xs->data & PGOFSET;
      pendp->vm_link_data.pages  = round_page(xs->data+xs->datalen-sva)/NBPG;
	    
      for(n=0; n<pendp->vm_link_data.pages; n++)
	pendp->vm_link_data.pa[n] = kvtop(sva + n*NBPG);
    }
#endif

  if ((dev->sc_flags & FAS_ACTIVE) || 
      (dev->sc_nexus[target].flags & FAS_UNIT_BUSY))
    {
      s = splbio();
      TAILQ_INSERT_TAIL(&dev->sc_xs_pending, pendp, link);
      splx(s);
    }
  else
    fas_donextcmd(dev, pendp);

  if (flags & SCSI_POLL)
    return(COMPLETE);
  return(SUCCESSFULLY_QUEUED);
}

fas_donextcmd(struct fas_softc *dev, struct fas_pending *pendp)
{
  struct scsi_xfer	*xs;
  int			 stat, s;

  xs = pendp->xs;

  if (pendp->xs->flags & SCSI_RESET)
    fasreset(dev, 2);

  if (pendp->xs->flags & SCSI_POLL)
    {
      s = splbio();
      stat = fasicmd(dev, pendp);
      TAILQ_INSERT_TAIL(&dev->sc_xs_free, pendp, link);
      splx(s);
      fas_scsidone(dev, xs, stat);
    }
  else
    {
      fasgo(dev, pendp);
      return;
    }
}

void fas_scsidone(struct fas_softc *dev, struct scsi_xfer *xs, int stat)
{
  struct fas_pending	*pendp;
  int			 s;

  if (xs == NULL)
    panic("fas_scsidone: no xs!");

  xs->status = stat;

  if (stat == 0)
    xs->resid = 0;
  else
    {
      switch(stat)
	{
	case SCSI_CHECK:
	  xs->error = XS_SENSE;
	  break;
	case SCSI_BUSY:
	  xs->error = XS_BUSY;
	  break;
	default:
	  xs->error = XS_TIMEOUT;
	  QPRINTF(("fas_scsicmd() bad %x\n", stat));
	  break;
	}
    }

  xs->flags |= ITSDONE;

  s = splbio();
  pendp = dev->sc_xs_pending.tqh_first;
  while(pendp)
    {
      if (!(dev->sc_nexus[pendp->xs->sc_link->target].flags & FAS_UNIT_BUSY))
	break;
      pendp = pendp->link.tqe_next;
    }

  if (pendp)
    {
      TAILQ_REMOVE(&dev->sc_xs_pending, pendp, link);
    }

  splx(s);
  scsi_done(xs);

  if (pendp)
    fas_donextcmd(dev, pendp);
}

void fasreset(struct fas_softc *dev, int how)
{
  fas_regmap_p	rp;
  int		i;

  for(i=0; i<8; i++)
    fas_init_nexus(dev, &dev->sc_nexus[i]);

  rp = dev->sc_fas;

  if (how & 1)
    {
      *rp->fas_command = FAS_CMD_RESET_CHIP;
      delay(1);
      *rp->fas_command = FAS_CMD_NOP;

      *rp->fas_config1 = dev->sc_config1;
      *rp->fas_config2 = dev->sc_config2;
      *rp->fas_config3 = dev->sc_config3;
      *rp->fas_timeout = dev->sc_timeout_val;
      *rp->fas_clkconv = dev->sc_clock_conv_fact & FAS_CLOCK_CONVERSION_MASK;
    }
  if (how & 2)
    {
      *rp->fas_command = FAS_CMD_RESET_SCSI_BUS;
      delay(10000);

/* Skip interrupt generated by RESET_SCSI_BUS */
      while(*rp->fas_status & FAS_STAT_INTERRUPT_PENDING)
	{
	  dev->sc_status = *rp->fas_status;
	  dev->sc_interrupt = *rp->fas_interrupt;
	  
	  delay(10000);
	}

      dev->sc_status = *rp->fas_status;
      dev->sc_interrupt = *rp->fas_interrupt;
    }

  delay(250000);		/* RESET to SELECT DELAY */
}

void fas_save_pointers(struct fas_softc *dev)
{
  struct nexus	*nx;

  nx = dev->sc_cur_nexus;
  if (nx)
    {
      nx->cur_link	= dev->sc_cur_link;
      nx->max_link	= dev->sc_max_link;
      nx->buf		= dev->sc_buf;
      nx->len		= dev->sc_len;
      nx->dma_len	= dev->sc_dma_len;
      nx->dma_buf	= dev->sc_dma_buf;
      nx->dma_blk_flg	= dev->sc_dma_blk_flg;
      nx->dma_blk_len	= dev->sc_dma_blk_len;
      nx->dma_blk_ptr	= dev->sc_dma_blk_ptr;
    }
}

void fas_restore_pointers(struct fas_softc *dev)
{
  struct nexus	*nx;

  nx = dev->sc_cur_nexus;
  if (nx)
    {
      dev->sc_cur_link	  = nx->cur_link;
      dev->sc_max_link	  = nx->max_link;
      dev->sc_buf	  = nx->buf;
      dev->sc_len	  = nx->len;
      dev->sc_dma_len	  = nx->dma_len;
      dev->sc_dma_buf	  = nx->dma_buf;
      dev->sc_dma_blk_flg = nx->dma_blk_flg;
      dev->sc_dma_blk_len = nx->dma_blk_len;
      dev->sc_dma_blk_ptr = nx->dma_blk_ptr;
      dev->sc_chain	  = nx->dma;
      dev->sc_unit	  = (nx->lun_unit & 0x0F);
      dev->sc_lun	  = (nx->lun_unit & 0xF0) >> 4;
    }
}

int fas_request_sense(struct fas_softc *dev, struct scsi_xfer *xs)
{
  struct fas_pending	 pend;
  struct scsi_sense	 rqs;
  int			 stat;

  /* Fake a fas_pending structure */
  pend.vm_link_data.pages = 0;
  pend.xs		  = xs;

  rqs.opcode = REQUEST_SENSE;
  rqs.byte2 = xs->sc_link->lun << 5;
#ifdef not_yet
  rqs.length = xs->req_sense_length ? xs->req_sense_length : sizeof(xs->sense);
#else
  rqs.length = sizeof(xs->sense);
#endif
  
  rqs.unused[0] = rqs.unused[1] = rqs.control = 0;

  if (dev->sc_cur_nexus->flags & FAS_IMMEDIATE)
    stat = fasselect(dev, &pend, (char *)&rqs, sizeof(rqs),
		     (char *)&xs->sense, rqs.length, FAS_SELECT_I);
  else
    stat = fasselect(dev, &pend, (char *)&rqs, sizeof(rqs),
		     (char *)&xs->sense, rqs.length, FAS_SELECT_RS);

  return(stat);
}

/* Immediate commands */
void fasiwait(struct fas_softc *dev)
{
  fas_regmap_p	rp;

  if (dev->sc_flags & FAS_DONT_WAIT)
    {
      dev->sc_flags &= ~FAS_DONT_WAIT;
      return;
    }

  rp = dev->sc_fas;

  while(!(*rp->fas_status & FAS_STAT_INTERRUPT_PENDING))
    delay(1);

  dev->sc_status = *rp->fas_status;
  dev->sc_interrupt = *rp->fas_interrupt;
  if (dev->sc_interrupt & FAS_INT_RESELECTED)
    {
      dev->sc_resel[0] = *rp->fas_fifo;
      dev->sc_resel[1] = *rp->fas_fifo;
    }
}

void fas_ixfer(struct fas_softc *dev)
{
  fas_regmap_p	 rp;
  u_char	*buf;
  int		 len, mode, phase;

  rp = dev->sc_fas;
  buf = dev->sc_buf;
  len = dev->sc_len;

  phase = dev->sc_status & FAS_STAT_PHASE_MASK;
  mode = (phase == FAS_PHASE_DATA_IN);

  while(len && ((dev->sc_status & FAS_STAT_PHASE_MASK) == phase))
    if (mode)
      {
	*rp->fas_command = FAS_CMD_TRANSFER_INFO;
	
	fasiwait(dev);
	
	*buf++ = *rp->fas_fifo;
	len--;
      }
    else
      {
	len--;
	*rp->fas_fifo = *buf++;
	*rp->fas_command = FAS_CMD_TRANSFER_INFO;

	fasiwait(dev);
      }

  dev->sc_buf = buf;
  dev->sc_len = len;
  dev->sc_flags |= FAS_DONT_WAIT;
}

void fas_build_sdtrm(struct fas_softc *dev, int period, int offset)
{
  dev->sc_msg_out[0] = 0x01;
  dev->sc_msg_out[1] = 0x03;
  dev->sc_msg_out[2] = 0x01;
  dev->sc_msg_out[3] = period/4;
  dev->sc_msg_out[4] = offset;
  dev->sc_msg_out_len = 5;
}

int fasselect(struct fas_softc *dev, struct fas_pending *pendp,
	      unsigned char *cbuf, int clen, 
	      unsigned char *buf, int len, int mode)
{
  fas_regmap_p	 rp;
  struct nexus	*nexus;
  int		 s;
  u_char	 ID, sync, cmd;
  short		 target, lun;

  rp = dev->sc_fas;

  if (dev->sc_config_flags & (FAS_NO_SYNCH | FAS_NO_DMA))
    mode &= ~FAS_SELECT_S;
  if (dev->sc_config_flags & FAS_NO_RESELECT)
    mode &= ~FAS_SELECT_R;

  s = splbio();

  target = pendp->xs->sc_link->target;
  lun    = pendp->xs->sc_link->lun;

  if ((dev->sc_flags & FAS_ACTIVE) ||
      (dev->sc_nexus[target].flags & FAS_UNIT_BUSY))
    {
      splx(s);
      return(0);
    }

  nexus = &dev->sc_nexus[target];
  ID    = ((mode & FAS_SELECT_R) ? 0xC0 : 0x80) | lun;

  dev->sc_flags |= FAS_ACTIVE;
  nexus->flags |= FAS_UNIT_BUSY;

  if (*rp->fas_status & FAS_STAT_INTERRUPT_PENDING)
    {
      nexus->flags &= ~FAS_UNIT_BUSY;
      splx(s);
      return(0);
    }

  *rp->fas_config1 = dev->sc_config1;
  *rp->fas_timeout = dev->sc_timeout_val;
  *rp->fas_dest_id = target;

  *rp->fas_fifo = ID;
  *rp->fas_command = FAS_CMD_SEL_ATN_STOP;

  fasiwait(dev);

  if (dev->sc_interrupt & FAS_INT_RESELECTED)
    {
      nexus->flags &= ~FAS_UNIT_BUSY;
      fasintr(dev);
      splx(s);
      return(0);
    }

  dev->sc_cur_nexus	= nexus;

  splx(s);

  nexus->clen = clen;
  bcopy(cbuf, nexus->cbuf, nexus->clen);
  nexus->cur_link	= 0;
  nexus->dma_len	= 0;
  nexus->dma_buf	= 0;
  nexus->dma_blk_len	= 0;
  nexus->dma_blk_ptr	= 0;
  nexus->len		= len;
  nexus->buf		= buf;
  nexus->xs		= pendp->xs;
#ifdef FAS_NEED_VM_PATCH
  nexus->vm_link_data	= pendp->vm_link_data;
#endif
  nexus->lun_unit	= (lun << 4) | target;
  nexus->state		= FAS_NS_SELECTED;
  nexus->flags	       &= ~(FAS_HAS_MSG		| FAS_SENSING
			  | FAS_GOT_DISCONN	| FAS_DISCONNECTED
			  | FAS_DO_SDTR		| FAS_SDTR_SENT
			  | FAS_IMMEDIATE	| FAS_DEBUG);

  if (nexus->flags & FAS_REQUEST_SENSE)
    {
      nexus->flags &= ~FAS_REQUEST_SENSE;
      nexus->flags |=  FAS_SENSING;
    }

  if (mode & FAS_SELECT_I)
    nexus->flags       |= FAS_IMMEDIATE;

  sync  = ((mode & FAS_SELECT_S) ? 1 : 0);

  if (sync && (mode & FAS_SELECT_I))
    sync = 0;

  if (!sync && ((nexus->flags & FAS_SYNC_TESTED) && (nexus->offset != 0)))
    {
      nexus->flags |= FAS_DO_SDTR;
      nexus->period = 200;
      nexus->offset = 0;
    }
  else if (sync && !(nexus->flags & FAS_SYNC_TESTED))
    {
      nexus->flags |= FAS_DO_SDTR;
      nexus->period = ((dev->sc_clock_freq>25) ? 100 : 200);
      nexus->offset = 8;

      if ((nexus->period == 100) && (dev->sc_config_flags & FAS_SLOW_CABLE))
	nexus->period = 200;
    }

  if (mode & FAS_SELECT_I)
    {
      nexus->dma[0].ptr = (vm_offset_t)buf;
      nexus->dma[0].len = len;
      nexus->dma[0].flg = FAS_CHAIN_PRG;
      nexus->max_link   = 1;
    }
  else
    {
#ifdef FAS_NEED_VM_PATCH
      if (nexus->vm_link_data.pages)
	fas_link_vm_link(dev, &nexus->vm_link_data);
#endif
      nexus->max_link = dev->sc_build_dma_chain(dev, nexus->dma, buf, len);
    }

  if (cpu040 && len && !(mode & FAS_SELECT_I))
    dma_cachectl(buf, len);

  fasintr(dev);

  return(1);
}

int fasgo(struct fas_softc *dev, struct fas_pending *pendp)
{
  int	 s;
  char	*buf;

  buf    = pendp->xs->data;

#ifdef FAS_NEED_VM_PATCH
  if (pendp->vm_link_data.pages)
    buf = dev->sc_vm_link + pendp->vm_link_data.offset;
#endif

  if (fasselect(dev, pendp, (char *)pendp->xs->cmd, pendp->xs->cmdlen,
		buf, pendp->xs->datalen, FAS_SELECT_RS))
    {
      s = splbio();
      TAILQ_INSERT_TAIL(&dev->sc_xs_free, pendp, link);
      splx(s);
    }
  else
    {
      s = splbio();
      TAILQ_INSERT_TAIL(&dev->sc_xs_pending, pendp, link);
      splx(s);
    }

  return(0);
}

int fas_pretests(struct fas_softc *dev, fas_regmap_p rp)
{
  struct nexus	*nexus;
  int		 i, s;

  if (dev->sc_interrupt & FAS_INT_SCSI_RESET_DETECTED)
    {
      dev->sc_flags &= ~FAS_RESEL_ENABLED;
      for(i=0; i<8; i++)
	{
	  if (dev->sc_nexus[i].xs)
	    fas_scsidone(dev, dev->sc_nexus[i].xs, -1);

	  fas_init_nexus(dev, &dev->sc_nexus[i]);
	}
      printf("fasintr: SCSI-RESET detected!");
      return(-1);
    }

  if (dev->sc_interrupt & FAS_INT_ILLEGAL_COMMAND)
    {
      dev->sc_flags &= ~FAS_RESEL_ENABLED;
      printf("FIFO:");
      while(*rp->fas_fifo_flags & FAS_FIFO_COUNT_MASK)
	printf(" %x", *rp->fas_fifo);
      printf("\n");

      printf("CMD: %x\n", rp->fas_command);
      panic("fasintr: ILLEGAL COMMAND!");
    }

  if (dev->sc_interrupt & FAS_INT_RESELECTED)
    {
      s = splbio();
      dev->sc_flags |= FAS_ACTIVE;
      dev->sc_flags &= ~FAS_RESEL_ENABLED;
      splx(s);

      if (dev->sc_units_disconnected)
	{
	  dev->sc_resel[0] &= ~(1<<dev->sc_host_id);

	  for(i=0; i<8; i++)
	    if (dev->sc_resel[0] & (1<<i))
	      break;

	  if (i == 8)
	    panic("Illegal reselection!");

	  if (dev->sc_nexus[i].flags & FAS_DISCONNECTED)
	    {
	      dev->sc_cur_nexus = &dev->sc_nexus[i];
	      nexus = dev->sc_cur_nexus;

	      *rp->fas_syncper = nexus->syncper;
	      *rp->fas_syncoff = nexus->syncoff;
	      *rp->fas_config3 = nexus->config3;

	      *rp->fas_dest_id = i & 7;

	      nexus->flags &= ~(FAS_DISCONNECTED | FAS_GOT_DISCONN);

	      dev->sc_units_disconnected--;
	      dev->sc_msg_in_len= 0;

#ifdef FAS_NEED_VM_PATCH
	      if (nexus->vm_link_data.pages)
		fas_link_vm_link(dev, &nexus->vm_link_data);
#endif

	      fas_restore_pointers(dev);

	      *rp->fas_command = FAS_CMD_MESSAGE_ACCEPTED;

	      return(1);
	    }
	}

      printf("fasintr: resel[0] %x resel[1] %x disconnected %d\n",
	     dev->sc_resel[0], dev->sc_resel[1], dev->sc_units_disconnected);
      panic("fasintr: Unexpected reselection!");
    }

  return(0);
}

int fas_midaction(struct fas_softc *dev, fas_regmap_p rp, struct nexus *nexus)
{
  int	i, left, len, s;

  if (dev->sc_interrupt & FAS_INT_DISCONNECT)
    {
      s = splbio();

      dev->sc_flags &= ~FAS_ACTIVE;

#ifdef FAS_NEED_VM_PATCH
      fas_unlink_vm_link(dev);
#endif

      switch(nexus->state)
	{
	case FAS_NS_SELECTED:
	  nexus->xs->resid = dev->sc_len;

     	  *rp->fas_command = FAS_CMD_FLUSH_FIFO;
	  nexus->status = -2;
	  nexus->flags &= ~FAS_UNIT_BUSY;
	  nexus->state = FAS_NS_FINISHED;
	  break;
	  
	case FAS_NS_SENSE:
	  dev->sc_led_off(dev);
	  nexus->flags |= FAS_REQUEST_SENSE;
	  nexus->flags &= ~FAS_UNIT_BUSY;
	  if (!fas_request_sense(dev, nexus->xs))
	    nexus->flags |= FAS_REQUEST_SENSE | FAS_UNIT_BUSY;
	  break;
	  
	case FAS_NS_DONE:
	  nexus->xs->resid = dev->sc_len;

	  nexus->flags &= ~FAS_UNIT_BUSY;
	  nexus->state  = FAS_NS_FINISHED;
	  dev->sc_led_off(dev);
	  break;
	  
	default:
	  if (nexus->flags & FAS_GOT_DISCONN)
	    {
	      *rp->fas_command = FAS_CMD_FLUSH_FIFO;
	      
	      nexus->state  = FAS_NS_DISCONNECTED;
	      nexus->flags &= ~FAS_GOT_DISCONN;
	      nexus->flags |= FAS_DISCONNECTED;

	      dev->sc_units_disconnected++;
	    }
	  else
	    {
	      printf("fasintr: Unexpected disconnection\n");
	      printf("fasintr: unit %x state %d phase %d flags %x cmd %x\n",
		     nexus->lun_unit, nexus->state, 
		     dev->sc_status & FAS_STAT_PHASE_MASK,
		     nexus->flags, nexus->cbuf[0]);
	      
	      printf("FIFO:");
	      while(*rp->fas_fifo_flags & FAS_FIFO_COUNT_MASK)
		printf(" %x", *rp->fas_fifo);
	      printf("\n");

	      nexus->xs->resid = dev->sc_len;

	      nexus->flags &= ~FAS_UNIT_BUSY;
	      nexus->state = FAS_NS_FINISHED;
	      nexus->status = -3;
	      
	      dev->sc_led_off(dev);
	    }
	  break;
	}

      if (dev->sc_units_disconnected && !(dev->sc_flags & FAS_ACTIVE))
	{
	  dev->sc_flags |= FAS_RESEL_ENABLED;
	  *rp->fas_command = FAS_CMD_ENABLE_RESEL;
	}

      splx(s);

      for(i=0; i<8; i++)
	if (dev->sc_nexus[i].flags & FAS_REQUEST_SENSE)
	  {
	    dev->sc_nexus[i].flags &= ~FAS_UNIT_BUSY;
	    if (!fas_request_sense(dev, dev->sc_nexus[i].xs))
	      dev->sc_nexus[i].flags |= FAS_REQUEST_SENSE | FAS_UNIT_BUSY;
	    else
	      break;
	  }
      
      if (nexus->state == FAS_NS_FINISHED)
	{
	  if (!(nexus->flags & FAS_IMMEDIATE))
	    fas_scsidone(dev, nexus->xs, nexus->status);
	  return(-1);
	}
      else
	return(1);
    }

  switch(nexus->state)
    {
    case FAS_NS_SELECTED:
      *rp->fas_syncper = nexus->syncper;
      *rp->fas_syncoff = nexus->syncoff;
      *rp->fas_config3 = nexus->config3;

      dev->sc_flags &= ~FAS_RESEL_ENABLED;

      fas_restore_pointers(dev);

      if (!(nexus->flags & FAS_SENSING))
	nexus->status	= 0xFF;
      dev->sc_msg_in[0] = 0xFF;
      dev->sc_msg_in_len= 0;

      dev->sc_led_on(dev);
      break;

    case FAS_NS_DATA_IN:
    case FAS_NS_DATA_OUT:
      if (dev->sc_dma_len)
	{
	  if (dev->sc_cur_link < dev->sc_max_link)
	    {
	      left = dev->sc_setup_dma(dev, 0, 0, FAS_DMA_CLEAR);
	      len  = dev->sc_dma_len;

	      if (nexus->state == FAS_NS_DATA_IN)
		{
		  if (dev->sc_dma_buf == dev->sc_bump_pa)
		    {
		      while((*rp->fas_fifo_flags & FAS_FIFO_COUNT_MASK)&& left)
			dev->sc_bump_va[len-(left--)] = *rp->fas_fifo;
		    
		      bcopy(dev->sc_bump_va, dev->sc_buf, len-left);
		    }
		}
	      else
		{
		  left += *rp->fas_fifo_flags & FAS_FIFO_COUNT_MASK;
		  *rp->fas_command = FAS_CMD_FLUSH_FIFO;

		  if (dev->sc_dma_buf == dev->sc_bump_pa)
		    {
		      if (left)
			printf("FAS_MIDACTIPON: left in data_out dump\n");
		    }
		}
	    
	      dev->sc_len -= len-left;
	      dev->sc_buf += len-left;
	    
	      dev->sc_dma_buf += len-left;
	      dev->sc_dma_len  = left;
	    
	      dev->sc_dma_blk_ptr += len-left;
	      dev->sc_dma_blk_len -= len-left;
	    
	      if (!dev->sc_dma_blk_len)
		dev->sc_cur_link++;
	    }
	}
      break;

    case FAS_NS_STATUS:
      if (!(nexus->flags & FAS_SENSING))
	nexus->status = *rp->fas_fifo;
      else
	if (*rp->fas_fifo != 0)
	  panic("REQUEST SENSE RETURNED BAD STATUS!");
      break;

    default:
      break;
    }

  return(0);
}

int fas_postaction(struct fas_softc *dev, fas_regmap_p rp, struct nexus *nexus)
{
  int		i, left, len;
  u_char	cmd;
  short		offset, period;

  cmd = 0;

  switch(dev->sc_status & FAS_STAT_PHASE_MASK)
    {
    case FAS_PHASE_DATA_OUT:
    case FAS_PHASE_DATA_IN:
      if ((dev->sc_status & FAS_STAT_PHASE_MASK) == FAS_PHASE_DATA_OUT)
	nexus->state = FAS_NS_DATA_OUT;
      else
	nexus->state = FAS_NS_DATA_IN;

      dev->sc_setup_dma(dev, 0, 0, FAS_DMA_CLEAR);
      if (dev->sc_cur_link < dev->sc_max_link)
	{
	  if (!dev->sc_dma_blk_len)
	    {
	      dev->sc_dma_blk_ptr = dev->sc_chain[dev->sc_cur_link].ptr;
	      dev->sc_dma_blk_len = dev->sc_chain[dev->sc_cur_link].len;
	      dev->sc_dma_blk_flg = dev->sc_chain[dev->sc_cur_link].flg;
	    }

	  if (dev->sc_dma_blk_flg == FAS_CHAIN_PRG)
	    {
	      fas_ixfer(dev);
	      dev->sc_cur_link++;
	      dev->sc_dma_len = 0;
	      break;
	    }
	  else if (dev->sc_dma_blk_flg == FAS_CHAIN_BUMP)
	    len = dev->sc_dma_blk_len;
	  else
	    len=dev->sc_need_bump(dev,
				  dev->sc_dma_blk_ptr,
				  dev->sc_dma_blk_len);

	  if (len)
	    {
	      dev->sc_dma_buf = dev->sc_bump_pa;
	      dev->sc_dma_len = len;
	      
	      if (nexus->state == FAS_NS_DATA_OUT)
		bcopy(dev->sc_buf, dev->sc_bump_va, dev->sc_dma_len);
	    }
	  else
	    {
	      dev->sc_dma_buf = dev->sc_dma_blk_ptr;
	      dev->sc_dma_len = dev->sc_dma_blk_len;
	    }

	  if (nexus->state == FAS_NS_DATA_OUT)
	    dev->sc_setup_dma(dev, dev->sc_dma_buf, dev->sc_dma_len,
			      FAS_DMA_WRITE);
	  else
	    dev->sc_setup_dma(dev, dev->sc_dma_buf, dev->sc_dma_len,
			      FAS_DMA_READ);

	  cmd = FAS_CMD_TRANSFER_INFO | FAS_CMD_DMA;
	}
      else
	{
	  *rp->fas_tc_low = 256;
	  *rp->fas_tc_mid = 0;
	  *rp->fas_tc_high = 0;
	  cmd = FAS_CMD_TRANSFER_PAD;
	}
      break;

    case FAS_PHASE_COMMAND:
      *rp->fas_command = FAS_CMD_FLUSH_FIFO;
      for(i=0; i<5; i++);
      nexus->state = FAS_NS_SVC;

      for(i=0; i<nexus->clen; i++)
	*rp->fas_fifo = nexus->cbuf[i] | (i==1 ? (dev->sc_lun<<5) : 0);
      cmd = FAS_CMD_TRANSFER_INFO;
      break;

    case FAS_PHASE_STATUS:
      nexus->state = FAS_NS_STATUS;
      cmd = FAS_CMD_COMMAND_COMPLETE;
      break;

    case FAS_PHASE_MESSAGE_OUT:
      nexus->state = FAS_NS_MSG_OUT;
      if (nexus->flags & FAS_DO_SDTR)
	{
	  fas_build_sdtrm(dev, nexus->period, nexus->offset);
	  nexus->flags |= FAS_SDTR_SENT;
	  nexus->flags &= FAS_DO_SDTR;
	}

      cmd = FAS_CMD_TRANSFER_INFO;
      if (dev->sc_msg_out_len)
	{
	  *rp->fas_command = FAS_CMD_FLUSH_FIFO;
	  for(i=0; i<5; i++);
	  for(i=0; i<dev->sc_msg_out_len; i++)
	    *rp->fas_fifo = dev->sc_msg_out[i];
	  dev->sc_msg_out_len = 0;
	}
      else
	*rp->fas_fifo = 8;	/* NOP */
      break;

    case FAS_PHASE_MESSAGE_IN:
      nexus->state = FAS_NS_MSG_IN;
      while(!(nexus->flags & FAS_HAS_MSG))
	{
	  *rp->fas_command = FAS_CMD_TRANSFER_INFO;
	  fasiwait(dev);

	  dev->sc_msg_in[dev->sc_msg_in_len++] = *rp->fas_fifo;

	  if (dev->sc_msg_in[0] >= 0x80)       ;
	  else if (dev->sc_msg_in[0] >= 0x30)  ;
	  else if (dev->sc_msg_in[0] >= 0x20)
	    {
	      if (dev->sc_msg_in_len == 2)
		{
		  nexus->flags |= FAS_HAS_MSG;
		  break;
		}
	    }
	  else if (dev->sc_msg_in[0] == 1)
	    {
	      if (dev->sc_msg_in_len >= 2)
		if ((dev->sc_msg_in[1]+2) == dev->sc_msg_in_len)
		  {
		    nexus->flags |= FAS_HAS_MSG;
		    break;
		  }
	    }
	  else
	    {
	      nexus->flags |= FAS_HAS_MSG;
	      break;
	    }

	  *rp->fas_command = FAS_CMD_MESSAGE_ACCEPTED;
	  fasiwait(dev);

	  if ((dev->sc_status & FAS_STAT_PHASE_MASK) != FAS_PHASE_MESSAGE_IN)
	    break;
	}

      cmd = FAS_CMD_MESSAGE_ACCEPTED;
      if (nexus->flags & FAS_HAS_MSG)
	{
	  switch(dev->sc_msg_in[0])
	    {
	    case 0x00:	/* COMMAND COMPLETE */
	      if ((nexus->status == SCSI_CHECK) &&
		  !(nexus->flags & FAS_SENSING))
		nexus->state = FAS_NS_SENSE;
	      else
		nexus->state = FAS_NS_DONE;
	      break;
	    case 0x04:	/* DISCONNECT */
	      nexus->flags |= FAS_GOT_DISCONN;
	      break;
	    case 0x02:	/* SAVE DATA POINTER */
	      fas_save_pointers(dev);
	      break;
	    case 0x03:	/* RESTORE DATA POINTERS */
	      fas_restore_pointers(dev);
	      break;
	    case 0x07:	/* MESSAGE REJECT */
	      if (nexus->flags & FAS_SDTR_SENT)
		{
		  nexus->flags &= ~FAS_SDTR_SENT;

		  nexus->config3 &= ~FAS_CFG3_FASTSCSI;
		  nexus->syncper = 5;
		  nexus->syncoff = 0;

		  *rp->fas_syncper = nexus->syncper;
		  *rp->fas_syncoff = nexus->syncoff;
		  *rp->fas_config3 = nexus->config3;
		}
	      else
		panic("fasintr: Unknown message rejected!");
	      break;
	    case 0x08:	/* MO OPERATION */
	      break;
	    case 0x01:	/* EXTENDED MESSAGE */
	      switch(dev->sc_msg_in[2])
		{
		case 0x00:	/* MODIFY DATA POINTERS */
		  break;
		case 0x01:	/* SYNCHRONOUS DATA TRANSFER REQUEST */
		  period = 4*dev->sc_msg_in[3];
		  offset = dev->sc_msg_in[4];

		  if (offset == 0)
		    period = 5*dev->sc_clock_period;

		  nexus->syncper = period/dev->sc_clock_period;
		  nexus->syncoff = offset;

		  if (period < 200)
		    nexus->config3 |= FAS_CFG3_FASTSCSI;
		  else
		    nexus->config3 &= ~FAS_CFG3_FASTSCSI;

		  nexus->flags |= FAS_SYNC_TESTED;

		  *rp->fas_syncper = nexus->syncper;
		  *rp->fas_syncoff = nexus->syncoff;
		  *rp->fas_config3 = nexus->config3;

		  nexus->flags &= ~FAS_SDTR_SENT;
		  break;
		case 0x02:	/* EXTENDED IDENTIFY (SCSI-1) */
		  break;
		case 0x03:	/* WIDE DATA TRANSFER REQUEST */
		  dev->sc_msg_out[0] = 0x07;
		  dev->sc_msg_out_len = 1;
		  *rp->fas_command = FAS_CMD_SET_ATN;
		  cmd = FAS_CMD_MESSAGE_ACCEPTED;
		  break;
		default:
		  break;
		}
	      break;
	    default:
	      dev->sc_msg_out[0] = 0x07;
	      dev->sc_msg_out_len = 1;
	      *rp->fas_command = FAS_CMD_SET_ATN;
	      cmd = FAS_CMD_MESSAGE_ACCEPTED;
	      break;
	    }
	  nexus->flags &= ~FAS_HAS_MSG;
	  dev->sc_msg_in_len = 0;
	}
      break;
    default:
      fasreset(dev, 2);
      printf("UNKNOWN PHASE! phase: %d\n",
	     dev->sc_status & FAS_STAT_PHASE_MASK);
      dev->sc_led_off(dev);
      if (!(nexus->flags & FAS_IMMEDIATE))
	fas_scsidone(dev, nexus->xs, -4);

      return(-1);
    }

  if (cmd)
    *rp->fas_command = cmd;

  return(0);
}

void fasintr(struct fas_softc *dev)
{
  fas_regmap_p	 rp;
  struct nexus	*nexus;
  int		 s;

  rp = dev->sc_fas;

  if (!fas_pretests(dev, rp))
    if (nexus = dev->sc_cur_nexus)
      if (!fas_midaction(dev, rp, nexus))
	fas_postaction(dev, rp, nexus);
}

int fasicmd(struct fas_softc *dev, struct fas_pending *pendp)
{
  fas_regmap_p	 rp;
  struct nexus	*nexus;

  rp = dev->sc_fas;

  if (!fasselect(dev, pendp, (char *)pendp->xs->cmd, pendp->xs->cmdlen,
		 (char *)pendp->xs->data, pendp->xs->datalen, FAS_SELECT_I))
    panic("fasicmd: Couldn't select unit");

  nexus = dev->sc_cur_nexus;

  while(nexus->state != FAS_NS_FINISHED)
    {
      fasiwait(dev);
      fasintr(dev);
    }

  nexus->flags &= ~FAS_SYNC_TESTED;

  return(nexus->status);
}
